/*
**    FILE    : stream.c   
**    PURPOSE : IO-functions
**    AUTHOR  : Bart De Canne, Belgium (bdc@barco.be)
**    COMPILER: BC++ 3.1 compiler
**    DEPENDENCIES:  Needs "galois.h" as an include file 
**    DATE    : 01-09-1994
**    VERSION : 1.0
*/

#define _STREAM
#include "galois.h"

#define DISK_BUFFER  4092

static int input_bit_count=0;
static ulong input_bit_buffer=0L;
static uint output_bit_count=0L;
static ulong output_bit_buffer=0L;
static int prev_length=0;
static byte buf[DISK_BUFFER];
static byte *buf_ptr;
static uint buf_full=0;
static FILE *fp_in,*fp_out;


/*************************************************************************/


void OpenInputStream(char *file) {

   BEGIN("OpenInputStream");
   if(NULL==(fp_in=fopen(file,"rb+"))) ERROR;
   length=filelength(fileno(fp_in));
   length<<=3;                                              /* in bits  */
   }

void CloseInputStream(void) {

   fclose(fp_in);
   }

void OpenOutputStream(char *file) {

   BEGIN("OpenOutputStream");
   if(NULL==(fp_out=fopen(file,"wb+"))) ERROR;
   }

void CloseOutputStream(void) {

   fclose(fp_out);
   }

/*************************************************************************/
/*
**    Resets all parameters regarding input and output buffers.
*/

void ResetBuffers(void) {

   input_bit_count=0;
   output_bit_count=0;
   input_bit_buffer=0L;
   output_bit_buffer=0L;
   buf_full=0;
   prev_length=0;
   }

/*************************************************************************/

bool EOFReached(void) {
   
   if(input_bit_count-prev_length<=0)
      return TRUE;
   return FALSE;
   }

/*************************************************************************/
/*
** Appends <code> of length <size> to output-stream.
** <code> should be right-aligned.
** <size> should be <= 16.
** NOTE: Don't forget to call PutBits() once more at the end of the
** output stream with <size> such that <output_bit_count>+<size> >= 8
** to make sure the last byte belonging to the stream is written to disk.
** (e.g. PutBits(0,7)).
*/

void PutBits(uint code,uint size) {

   output_bit_buffer |= (ulong) code << (32-size) >> output_bit_count;
   output_bit_count += size;
   
   while(output_bit_count >= 8){
      putc(output_bit_buffer >> 24,fp_out);
      output_bit_buffer <<= 8;
      output_bit_count -= 8;
      }
   }

/*************************************************************************/
/*
** Reads next 16 bits from the input stream. The input buffer is flushed
** by the value of <prev_length>. This value should initially be zero.
** The retrieved code is right-aligned.
*/

uint GetBits(uint size) {

   input_bit_buffer <<= prev_length;
   input_bit_count -= prev_length;
   while(input_bit_count <= 24 ) {
      if (!buf_full && !feof(fp_in)) {
         buf_full=fread(buf,1,DISK_BUFFER,fp_in);
         buf_ptr=buf;
         }
      if(buf_full){
         input_bit_buffer |= (ulong) *buf_ptr++ << 24 - input_bit_count;
         buf_full--;
         input_bit_count += 8;
         }
      else
         break;
      }
   prev_length=size;
   return(input_bit_buffer >> (32-size));
   }


/* 
** END OF STREAM.C
*/ 
